# -*- coding: utf-8 -*-
"""Module for calculating Zernike moments"""

import collections
import copy
import logging
import itertools
import math
import numbers
import functools
import random
from typing import Union, Tuple, Iterator

import numpy as np
from numpy import ma
from scipy import special

from . import base_moments
from . import functions
from . import mathutil
from . import geometric
from . import polynomials
from . import sph
from .utils import even, odd, inclusive

_LOGGER = logging.getLogger(__name__)

# pylint: disable=invalid-name

__all__ = 'ZernikeMoments', 'ZernikeMomentsCalculator', 'ZernikeMomentCalculator', 'from_deltas', 'from_gaussians'

NORMALISTAION = 3. / (4. * np.pi)


def from_deltas(
    n_max: int,
    positions: np.array,
    weights: Union[numbers.Number, np.array] = 1.,
    l_max: int = None
) -> 'ZernikeMoments':
    """Create a set of Zernike moments from a collection of delta functions with optional weights

    :param n_max: the maximum order of Zernike functions to use
    :param positions: the Cartesian positions of the delta functions
    :param weights: the weights of the delta functions, can be a scalar or np.array of the same
        length as positions
    :param l_max: the maximum spherical harmonic angular frequency, l, to go up to
    """
    _domain_check(positions)
    geom_moments = geometric.from_deltas(n_max, positions, weights)
    return from_geometric_moments(n_max, geom_moments, l_max)


def from_deltas_direct(
    nmax: int, positions: np.ndarray, weights: Union[numbers.Number, np.array] = 1., lmax=None
) -> 'ZernikeMoments':
    """This is the method used to calcualte fingerprints by AMP (at least when not using gradients).

    There is some discreptancy between this and the method from geometric moments proposed by Novotni and Klein, however
    it amounts to a simple rotation as the invariants are still the same.  We have this here, mainly so that
    we can easily compare with AMP code.
    """
    norm = NORMALISTAION
    zernike_moms = ZernikeMoments(n_max=nmax, l_max=lmax)

    if len(positions) > 0:
        spherical = mathutil.cart2sph(positions.T)  # Spherical coordinates
        for n in inclusive(nmax):
            for l in inclusive(n):
                if odd(n - l):
                    continue

                for m in inclusive(-l, l):
                    vals = norm * zernike_poly(n, l, m, spherical, condon_shortley=False).conjugate()
                    zernike_moms.array[n, l, m] = np.sum(weights * vals)
    else:
        zernike_moms.array.fill(0.)

    return zernike_moms


def from_gaussians(
    max_order: int,
    positions: np.ndarray,
    sigmas: Union[numbers.Number, np.array] = 0.1,
    weights: Union[numbers.Number, np.array] = 0.1
) -> 'ZernikeMoments':
    """Create a set of Zernike moments from a collection of Gaussian functions with the given sigmas
    and weights

    :param max_order: the order of Zernike moments to calculate to
    :param positions: the Cartesian positions of the delta functions
    :param sigmas: the sigmas of the Gaussians, can be a scalar or np.array of the same length
        as positions
    :param weights: the weights of the delta functions, can be a scalar or np.array of the same
        length as positions
    """
    _domain_check(positions)
    geom_moments = geometric.from_gaussians(max_order, positions, sigmas, weights)
    return from_geometric_moments(max_order, geom_moments)


def from_geometric_moments(
    n_max: int, geom_moments: Union[geometric.GeometricMoments, np.array], l_max: int = None
) -> 'ZernikeMoments':
    """Create a set of Zernike moments from a set of geometric moments

    :param n_max: the maximum order of Zernike functions to use
    :param geom_moments: the geometric moments
    :param l_max: the maximum spherical harmonic angular frequency, l, to go up to
    """
    return ZernikeMomentsCalculator(n_max, l_max)(geom_moments)


class ZernikeReconstructionQuery(base_moments.ReconstructionQuery):
    """A query object for Zernike moments that allows for faster reconstructions by caching the reconstruction grid"""

    def __init__(self, max_order: int, points: np.ndarray, moments: np.ndarray):
        super().__init__(max_order, points)
        self._moments = moments
        self._valid_idxs = None
        self._moments_in_domain = None

    @property
    def moments(self) -> np.ndarray:
        return self._moments

    @property
    def moments_in_domain(self) -> np.ndarray:
        if self._moments_in_domain is None:
            # Annoying indexing and unindexing, this returns a copy
            self._moments_in_domain = self._moments[:, :, :, self.valid_idxs][:, :, :, :, 0]
        return self._moments_in_domain

    @property
    def valid_idxs(self):
        """Get the indexes of points that are within the domain"""
        if self._valid_idxs is None:
            self._valid_idxs = _get_indices_in_domain(self.points)
        return self._valid_idxs


#: Helper tuple for working with Zernike moment indexes
ZernikeIndex = collections.namedtuple('ZernikeIndex', 'n l m')


class ZernikeMoments(base_moments.Moments):
    """A container class for calculated Zernike moments"""

    # pylint: disable=too-many-public-methods

    @classmethod
    def from_indexed(cls, indexed, max_order: int, dtype=None) -> 'ZernikeMoments':
        dtype = dtype or type(indexed[0, 0, 0])
        moments = cls(max_order, dtype=dtype)
        for (n, l, m) in moments.iter_indices():
            cls._set_moment(moments, n, l, m, indexed[n, l, m], set_minus_m=False)
        return moments

    @classmethod
    def from_vector(cls, n_max: int, vec: np.array, redundant=True) -> 'ZernikeMoments':
        moms = ZernikeMoments(n_max, dtype=vec.dtype)

        for (n, l, m), value in zip(iter_indices(redundant=redundant), vec):
            if n > n_max:
                break
            moms._set_moment(n, l, m, value, set_minus_m=not redundant)  # pylint: disable=protected-access

        return moms

    @classmethod
    def rand(cls, n_max: int, num_range=(-5., 5.)):
        moms = ZernikeMoments(n_max)
        for nlm in moms.iter_indices(redundant=False):
            mom = random.uniform(*num_range)
            if nlm.m != 0:
                # Only l != 0 moments have a complex part
                mom += +1j * random.uniform(*num_range)
            moms[nlm] = mom

        return moms

    def __init__(self, n_max: int, l_max=None, dtype=complex):
        """Construct a Zernike moments object
        """
        self._indices = index_traits(n_max, l_max)
        self._dtype = dtype

        moments = sph.create_array(self._indices, dtype=dtype)
        self._moments = moments
        # Fill it with NaN so that it's easy to spot any values that haven't been set but should be
        self._moments.fill(np.nan)

        # Create an index array that allows us to map from a multi-index like [5, 3, -1] to a linear index
        idx_array = ma.masked_array(np.empty(moments.shape, dtype=int), mask=moments.mask)
        idx_array[~moments.mask] = np.arange(moments.count())
        self._idx_array = idx_array

    def __eq__(self, other: 'ZernikeMoments'):
        """Test if another set of moments are equal to this one"""
        if not isinstance(other, ZernikeMoments):
            return False

        return np.all(self.vector == other.vector)

    def __add__(self, other: 'ZernikeMoments') -> 'ZernikeMoments':
        if not isinstance(other, ZernikeMoments):
            raise ValueError(f'Expected ZernikeMoments, got {type(other)}')

        moms = copy.deepcopy(self)
        moms.array.data[:] += other.array.data[:]

        return moms

    def __sub__(self, other: 'ZernikeMoments') -> 'ZernikeMoments':
        if not isinstance(other, ZernikeMoments):
            raise ValueError(f'Expected ZernikeMoments, got {type(other)}')

        moms = copy.deepcopy(self)
        moms.array.data[:] -= other.array.data[:]
        return moms

    @property
    def dtype(self):
        return self._dtype

    @property
    def array(self) -> np.ndarray:
        return self._moments

    def to_matrix(self):
        return self._moments

    @property
    def max_order(self) -> int:
        return self._indices.n.max

    @property
    def n_max(self) -> int:
        return self._indices.n.max

    @property
    def l_max(self) -> int:
        return self._indices.l.max

    def get_builder(self, mask: 'Optional[ZernikeMoments]' = None):
        return ZernikeMomentsBuilder(self.max_order, mask=mask)

    def get_mask(self) -> 'ZernikeMoments':
        moms = ZernikeMoments(self.n_max, self.l_max, dtype=object)
        moms.array.fill(None)
        return moms

    def randomise(self, num_range=(-5, 5), indices=(None, None, None)):
        for n, l, m in iter_indices_extended(indices, self.n_max, redundant=False):
            mom = random.uniform(*num_range)
            if m != 0:
                # Only l != 0 moments have a complex part
                mom += 1j * random.uniform(*num_range)
            self._set_moment(n, l, m, mom)

        return self

    @property
    def vector(self) -> np.ndarray:
        """Return this set of moments as a vector"""
        return self._moments.compressed()

    def __getitem__(self, item):
        return self._moments[item[0], item[1], item[2]]

    def __setitem__(self, key: Union[int, Tuple], value):
        self._set_moment(*key, value)

    def _set_moment(self, n: int, l: int, m: int, value, set_minus_m=True):
        """Set an individual moment.  This will automatically set the corresponding -m value"""
        if value is None:
            # Special case, usually happens if this is a mask
            self._moments[n, l, m] = None
            self._moments[n, l, -m] = None
        else:
            if m == 0:
                try:
                    if value.imag > 1e-9:
                        logging.warning(
                            'Trying to set value of moment %i,%i,%i to %d, '
                            'however these moment should always be real so discarding imaginary', n, l, m, value
                        )
                    value = np.real(value)
                except AttributeError:
                    pass
            elif set_minus_m:
                # Set the -m counterpart directly
                self._moments[n, l, -m] = (-1)**m * value.conjugate()

            # This will be reached for all m values
            self._moments[n, l, m] = value

    def iter(self, redundant=True):
        """Iterate tuples containing the moment indices and value"""
        for indices in self.iter_indices(redundant=redundant):
            yield indices, self[indices]

    @property
    def indices(self):
        return self._indices

    def iter_indices(self, redundant=True):
        yield from iter_indices(n_max=self.n_max, l_max=self.l_max, redundant=redundant)

    def linear_index(self, index: base_moments.Index) -> int:
        """Given a triple of Zernike moment indices this will return the corresponding linearly ordered index"""
        return self._idx_array[index]

    @staticmethod
    def num_moments(max_order: int, redundant=True) -> int:
        """Get the total number of Zernike moments up to the maximum order"""
        return sum(1 for _ in iter_indices(max_order, redundant=redundant))

    def moment(self, n: int, l: int, m: int) -> complex:
        """Get the n, l, m^th moment"""
        return self._moments[n, l, m]

    def value_at(self, x: np.array, order: int = None) -> float:
        """Reconstruct the value at x from the moments

        :param x: the point to get the value at
        :param order: the maximum order to go up to (defaults to the order of these moments)
        """
        order = order or self.n_max
        return self.reconstruct(self.create_reconstruction_query(x, order), order, zero_outside_domain=False)

    def reconstruct(self,
                    query: ZernikeReconstructionQuery,
                    order=None,
                    zero_outside_domain=True) -> Union[float, np.ndarray]:
        """Given a reconstruction query this will return the values of the function at the specified grid values"""
        order = order if order is not None else query.max_order
        moments = query.moments

        if len(query.points.shape) > 1:
            values = np.zeros(query.points.shape[0], dtype=self.dtype)
        else:
            values = 0.

        if zero_outside_domain:
            # Use only reconstruction moments from within the domain
            moments = query.moments_in_domain
            calculated_values = np.zeros(query.valid_idxs.size, dtype=self.dtype)
        else:
            calculated_values = values

        for n, l, m in iter_indices(n_max=order, redundant=False):
            omega = self._moments[n, l, m]
            z = ZernikeMomentsCalculator.change_of_basis(n, l, m)(moments)
            calculated_values += omega * z

            if m != 0:
                # Now do the symmetric -m part
                z_conj = (-1)**m * z.conjugate()
                calculated_values += z_conj * self._moments[n, l, -m]

        if zero_outside_domain:
            np.put(values, query.valid_idxs, calculated_values)

        return values.real

    def visualise(
        self,
        viewer='plotly_widget',
        query: ZernikeReconstructionQuery = None,
        max_order=None,
        num_points=31,
        square=False
    ):
        # pylint: disable=too-many-locals

        if viewer.startswith('plotly_widget'):
            volume_plot = viewer.endswith('volume')

            # Only do imports here so some of these tools aren't needed for the rest of milad
            import ipywidgets
            import plotly.graph_objects as go
            from milad import plotlytools

            if query is None:
                grid_points = self.get_grid(num_points, restrict_to_domain=False)
                query = self.create_reconstruction_query(grid_points, self.max_order)
            else:
                grid_points = query.points

            grid_values = self.reconstruct(query, max_order, zero_outside_domain=True)
            if square:
                grid_values = grid_values**2

            default_isomin = (grid_values.max() - grid_values.min()) * 0.6 + grid_values.min()
            default_isomax = grid_values.max()

            if volume_plot:
                data = plotlytools.volume(
                    grid_points,
                    grid_values,
                    isomin=grid_values.mean(),
                    opacity=0.1,
                    autocolorscale=True,
                )
            else:
                data = plotlytools.isosurface(
                    grid_points,
                    grid_values,
                    isomin=grid_values.mean(),
                    opacity=0.1,
                    autocolorscale=True,
                )

            fig = go.Figure(data=[data])
            widget = go.FigureWidget(fig)

            default_opacity = 0.05 if volume_plot else 0.4

            # pylint: disable=unused-variable
            @ipywidgets.interact(
                opacity=(0., 1., 0.01),
                isomin=(grid_values.min(), grid_values.max(), 0.5),
                isomax=(grid_values.min(), grid_values.max(), 0.5)
            )
            def update(opacity=default_opacity, isomin=default_isomin, isomax=default_isomax):
                with widget.batch_update():
                    isosurface = widget.data[0]
                    isosurface.opacity = opacity
                    isosurface.isomin = isomin
                    isosurface.isomax = isomax

            return widget

        raise ValueError(f"Unknown viewer '{viewer}'")

    @classmethod
    def create_reconstruction_query(cls, points: np.ndarray, max_order: int) -> ZernikeReconstructionQuery:
        """Create a query object that can be passed to self.value_at to reconstruct the Zernike functions
        at the given query points.  This is useful if the same point (or grid of points) is used multiple
        times as this precomputation can be done once and reused.

        :param points: the point to get the value at
        :param max_order: the maximum order to go up to (defaults to the order of these moments)
        """
        return ZernikeReconstructionQuery(max_order, points, cls._get_geometric_moments(points, max_order))

    @classmethod
    def _get_geometric_moments(cls, point: np.array, order: int) -> np.ndarray:
        """Get geometric moments from a point or set of points.  Used in reconstruction."""
        if len(point.shape) == 2:
            moms = [geometric.from_deltas(order, [pt]).to_matrix() for pt in point]
            moments = np.empty(list(moms[0].shape) + [len(moms)])
            for idx, entry in enumerate(moms):
                moments[:, :, :, idx] = entry
        else:
            moments = geometric.from_deltas(order, [point])

        return moments


def _get_indices_in_domain(points: np.ndarray) -> np.ndarray:
    """Calculate the lengths squared and get the corresponding indexes"""
    length_sq = (points**2).sum(axis=1)
    return np.argwhere(length_sq < 1)


class ZernikeMomentsBuilder(functions.Function):
    """This builder takes a vector of real values and constructs a set of complex Zernike moments."""
    input_type = np.ndarray
    supports_jacobian = True

    def __init__(self, n_max: int, l_max: int = None, mask: ZernikeMoments = None):
        super().__init__()
        self._n_max = n_max
        self._l_max = n_max if l_max is None else l_max
        self._mask = mask
        self._indices = None

    def apply_mask(self, moms: ZernikeMoments):
        """Given a set of moments this will set any values present from the mask"""
        if self._mask is None:
            raise RuntimeError('Cannot apply mask as, mask is None')
        for zernike_index, value in self._mask.iter(redundant=False):
            if value is not None:
                moms[zernike_index] = value

    @property
    def inverse(self) -> 'ZernikeMomentsBuilder.Inverse':
        return ZernikeMomentsBuilder.Inverse(self)

    def evaluate(self, state: np.ndarray, *, get_jacobian=False):  # pylint: disable=arguments-differ
        indices_list = tuple(self.iter_indices())
        if len(state) != len(indices_list):
            raise ValueError(
                f'Got input vector of length {len(state)} but was expecting one of length {len(indices_list)}'
            )

        # Create the moments and optionally a Jacobian
        moms = ZernikeMoments(self._n_max)
        moms.array.fill(0.)
        if self._mask is not None:
            self.apply_mask(moms)
        jac = np.zeros((len(moms), len(indices_list)), dtype=complex) if get_jacobian else None

        # Let's create the complex moments
        for idx, (zernike_idx, num_type) in enumerate(indices_list):
            moms[zernike_idx] += num_type * state[idx]

            if get_jacobian:
                n, l, m = zernike_idx
                lin_idx = moms.linear_index(zernike_idx)
                jac[lin_idx, idx] += 1 * num_type  # pylint: disable=unsupported-assignment-operation
                if zernike_idx.m > 0:
                    lin_idx = moms.linear_index((n, l, -m))
                    # pylint: disable=unsupported-assignment-operation
                    jac[lin_idx, idx] += (-1)**m * (1 if num_type == 1. else -1j)

        if get_jacobian:
            return moms, jac

        return moms

    def iter_indices(self):
        # Skip all the ones that are masked
        if self._indices is None:
            # Lazily create the indices
            self._indices = tuple(self._iter_indices(self._n_max, self._l_max))

        for zernike_index, num_type in self._indices:
            if self._mask is None or self._mask[zernike_index] is None:
                yield zernike_index, num_type

    @classmethod
    def _iter_indices(cls, n_max: int, l_max: int):
        # First deal with m = 0 as these are all real
        indices = index_traits(n_max, l_max)

        for n in inclusive(0, n_max):
            for l in indices.iter_l(n):
                if not even(n - l):
                    continue

                yield ZernikeIndex(n, l, 0), 1.

        # Now deal with m > 0, these have both a real and imaginary part
        for num_type in (1., 1j):
            for n in inclusive(n_max):
                for l in indices.iter_l(n):
                    if not even(n - l):
                        continue

                    for m in inclusive(1, l, 1):
                        yield ZernikeIndex(n, l, m), num_type

    class Inverse(functions.Function):
        """Turn a set a Zernike moments into a vector"""
        supports_jacobian = False

        def __init__(self, forward: 'ZernikeMomentsBuilder'):
            super().__init__()
            self._forward = forward

        @property
        def inverse(self) -> 'ZernikeMomentsBuilder':
            return self._forward

        def evaluate(  # pylint: disable=unused-argument
                self, state: ZernikeMoments, *, get_jacobian=False) -> np.ndarray:
            moms = []
            for indices, num_type in self._forward.iter_indices():
                mom = state[indices]
                moms.append(mom.real if num_type == 1. else mom.imag)

            return np.array(moms)


class ZernikeMomentsCalculator(base_moments.MomentsCalculator):
    """Calculate Zernike moments.

    Takes as input geometric moments or any state vector that is a valid input to GeometricMomentsCalculator
    """
    output_type = ZernikeMoments
    supports_jacobian = True
    dtype = complex
    input_type = geometric.GeometricMoments, np.ndarray, functions.Features
    _change_of_basis_polys = dict()  # Change of basis polynomials (geometric -> spherical harmonic)

    @classmethod
    def change_of_basis(cls, n: int, l: int, m: int) -> polynomials.HomogenousPolynomial:
        try:
            return cls._change_of_basis_polys[(n, l, m)]
        except KeyError:
            poly = sum_chi_nlm(n, l, m, polynomials.PolyBuilder())
            cls._change_of_basis_polys[(n, l, m)] = poly

            if m != 0 and (n, l, -m) not in cls._change_of_basis_polys:
                # Cache the -m version as well
                cls._change_of_basis_polys[(n, l, -m)] = (-1)**m * poly.conjugate()

            return poly

    def __init__(self, n_max: int, l_max: int = None, use_direct=False):
        super().__init__()
        self._n_max = n_max
        self._l_max = l_max
        self._geometric_moments_calculator = geometric.GeometricMomentsCalculator(n_max)
        self._jacobian = None
        self._use_direct = use_direct

    def evaluate(self,
                 state: functions.State,
                 *,
                 get_jacobian=False) -> Union[ZernikeMoments, Tuple[ZernikeMoments, np.ndarray]]:
        if self._use_direct:
            if not isinstance(state, functions.Features) or not all(
                isinstance(feature, functions.WeightedDelta) for feature in state
            ):
                raise ValueError('Can only use direct method on a set of features that are delta functions')

            return from_deltas_direct(
                self._n_max,
                positions=np.array([feature.pos for feature in state]),
                weights=np.array([feature.weight for feature in state]),
                lmax=self._l_max
            )

        if isinstance(state, geometric.GeometricMoments):
            geom_moments = state
            geom_jac = np.eye(state.size)
        else:
            # if not get_jacobian and isinstance(state, functions.Features) and \
            #         all(isinstance(feature, functions.WeightedDelta) for feature in state):
            #     positions = np.array([feature.pos for feature in state])
            #     weights = np.array([feature.weight for feature in state])
            #     return from_deltas_direct(self._n_max, positions, weights, lmax=self._l_max)

            geom_moments = self._geometric_moments_calculator(state, get_jacobian)
            if get_jacobian:
                geom_moments, geom_jac = geom_moments

        moments = ZernikeMoments(self._n_max, self._l_max, dtype=object if geom_moments.dtype == object else complex)

        # Get the moments themselves from polynomials of geometric moments
        for (n, l, m) in moments.iter_indices(redundant=False):
            moments[n, l, m] = omega_nl_m(n, l, m, geom_moments.array)

        if get_jacobian:
            jac = np.matmul(self._get_jacobian(), geom_jac)
            return moments, jac

        return moments

    def _get_jacobian(self):
        if self._jacobian is None:
            # Calculate the first time
            self._jacobian = get_jacobian_wrt_geom_moments(self._n_max, redundant=True)

        return self._jacobian

    def create_random(self, max_order: int = None) -> ZernikeMoments:
        max_order = max_order or self._n_max
        return ZernikeMoments.rand(max_order)


ZernikeMomentCalculator = ZernikeMomentsCalculator
"""An alias left here for compatibility reasons"""


def rand(max_order: int) -> ZernikeMoments:
    return ZernikeMomentsCalculator(max_order).create_random()


def get_jacobian_wrt_geom_moments(max_order: int, redundant=True):
    """Get the Jacobian of the Zernike moments wrt Geometric moments as input"""

    # Calculate
    class DerivativeTracker:

        def __init__(self, max_ord):
            self._max_order = max_ord
            self._num_coeffs = geometric.GeometricMoments.num_moments(max_ord)

        def __getitem__(self, item: base_moments.Index):
            if not isinstance(item, tuple):
                raise TypeError(item)

            vector = np.zeros(self._num_coeffs)
            try:
                vector[geometric.linear_index(self._max_order, item)] = 1.
            except IndexError:
                raise IndexError(f'index {item} out of bounds for geometric moments of order {self._max_order}')
            return vector

    tracker = DerivativeTracker(max_order)
    num_moments = ZernikeMoments.num_moments(max_order, redundant=redundant)  # Number of zernike moments
    num_geometric_moments = geometric.GeometricMoments.num_moments(max_order)
    jacobian = np.zeros((num_moments, num_geometric_moments), dtype=complex)

    cache = {}  # Used for negative m coefficients
    for idx, (n, l, m) in enumerate(iter_indices(max_order, redundant=redundant)):
        if redundant and m < 0:
            jacobian[idx, :] = (-1)**(-m) * cache.pop((n, l, -m)).conjugate()
        else:
            vec = omega_nl_m(n, l, m, tracker)
            jacobian[idx, :] = vec
            if redundant and m > 0:
                # Because of the iteration order (m=0 to l followed by m=-l to -1) the positives will always be
                # calculated first, so cache these for when we get around to the positive m's
                cache[(n, l, m)] = vec

    return jacobian


@functools.lru_cache(maxsize=None)
def factorial(n: int) -> int:
    return special.factorial(n, exact=True)


@functools.lru_cache(maxsize=None)
def binomial_coeff(n: int, k: int) -> int:
    return special.comb(n, k, exact=False)


@functools.lru_cache(maxsize=None)
def c_l_m(l: int, m: int) -> float:
    """Calculate the normalisation factor"""
    if m < 0:
        # make use of symmetry c_l^m == c_l^-m (no conjugate because these are real)
        return c_l_m(l, -m)

    return math.sqrt((2 * l + 1) * factorial(l + m) * factorial(l - m)) / factorial(l)


@functools.lru_cache(maxsize=None)
def q_kl_nu(k: int, l: int, nu: int) -> float:
    """Calculate the Zernike geometric moment conversion factors"""
    return (-1) ** k / 2 ** (2 * k) * \
           np.sqrt((2 * l + 4 * k + 3) / 3) * \
           binomial_coeff(2 * k, k) * \
           (-1) ** nu * \
           (binomial_coeff(k, nu) * binomial_coeff(2 * (k + l + nu) + 1, 2 * k)) / \
           binomial_coeff(k + l + nu, k)


def sum_chi_nlm(n: int, l: int, m: int, geom_moments: base_moments.MomentsProto) -> Union[complex, np.array]:
    """This call performs the change of basis transformation from a polynomial basis to Zernike"""
    assert_valid(n, l, m)
    if m < 0:
        # Symmetry relation
        return (-1)**(-m) * sum_chi_nlm(n, l, -m, geom_moments).conjugate()

    return c_l_m(l, m) * 2**(-m) * sum1(n=n, l=l, m=m, geom_moments=geom_moments)


def sum1(n: int, l: int, m: int, geom_moments: base_moments.MomentsProto) -> Union[complex, np.array]:
    """Calculate the Zernike geometric moment conversion factors"""
    k = int(math.floor((n - l) / 2))  # n - l is always even
    total = 0.
    for nu in inclusive(k):
        total += q_kl_nu(k, l, nu) * sum2(n=n, l=l, m=m, nu=nu, geom_moments=geom_moments)

    return total


def sum2(n: int, l: int, m: int, nu: int, geom_moments: base_moments.MomentsProto) -> Union[complex, np.array]:
    total = 0.
    for alpha in inclusive(nu):
        total += binomial_coeff(nu, alpha) * \
                 sum3(n=n, l=l, m=m, nu=nu, alpha=alpha, geom_moments=geom_moments)
    return total


def sum3(n: int, l: int, m: int, nu: int, alpha: int,
         geom_moments: base_moments.MomentsProto) -> Union[complex, np.array]:
    total = 0.
    for beta in inclusive(nu - alpha):
        total += binomial_coeff(nu - alpha, beta) * \
                 sum4(n=n, l=l, m=m, nu=nu, alpha=alpha, beta=beta, geom_moments=geom_moments)
    return total


def sum4(n: int, l: int, m: int, nu: int, alpha: int, beta: int,
         geom_moments: base_moments.MomentsProto) -> Union[complex, np.array]:
    total = 0.
    for u in inclusive(m):
        total += (-1) ** (m - u) * binomial_coeff(m, u) * 1j ** u * \
                 sum5(n=n, l=l, m=m, nu=nu, alpha=alpha, beta=beta, u=u,
                      geom_moments=geom_moments)
    return total


def sum5(n: int, l: int, m: int, nu: int, alpha: int, beta: int, u: int,
         geom_moments: base_moments.MomentsProto) -> Union[complex, np.array]:
    total = 0.
    for mu in inclusive(int(math.floor((l - m) / 2.))):
        total += (-1) ** mu * 2 ** (-2 * mu) * binomial_coeff(l, mu) * \
                 binomial_coeff(l - mu, m + mu) * \
                 sum6(n=n, l=l, m=m, nu=nu, alpha=alpha, beta=beta, u=u, mu=mu,
                      geom_moments=geom_moments)
    return total


def sum6(
    n: int, l: int, m: int, nu: int, alpha: int, beta: int, u: int, mu: int, geom_moments: base_moments.MomentsProto
) -> Union[complex, np.array]:
    total = 0.
    for v in inclusive(mu):
        r = 2 * (v + alpha) + u
        s = 2 * (mu - v + beta) + m - u
        t = 2 * (nu - alpha - beta - mu) + l - m

        if r + s + t > n:
            continue

        total += binomial_coeff(mu, v) * geom_moments[r, s, t]

    return total


def assert_valid(n: int, l: int, m: int):
    if not n >= 0:
        raise ValueError('n must be a positive integer')
    if not 0 <= l <= n:
        raise ValueError(f'l must be 0 <= l <= n, got n={n}, l={l}')
    if not -l <= m <= l:
        raise ValueError(f'm must be in the range -{l} <= m <= {l}, got {m}')
    if not even(n - l):
        raise ValueError('n - l must be even')


def omega_nl_m(n: int, l: int, m: int, geom_moments: base_moments.MomentsProto) -> complex:
    """Given a set of geometric moments this function will compute the corresponding Zernike moments
    """
    assert_valid(n, l, m)
    if m < 0:
        # Symmetry relation
        return (-1)**(-m) * omega_nl_m(n, l, -m, geom_moments).conjugate()

    return NORMALISTAION * ZernikeMomentsCalculator.change_of_basis(n, l, m)(geom_moments).conjugate()


def iter_indices(n_max: int = None, l_max: int = None, *, redundant=False) -> Iterator[ZernikeIndex]:
    """Iterate over Zernike function indices in lexicographic order.

    If redundant is True then all valid indices will be generated including those where there is a
    symmetry relation e.g.: o_22^-2 = (-1)^2 o22^2.

    :param n_max: order to provide indices up to, if None will generate all indices
    :param l_max: the maximum angular frequency to go up to
    :param redundant: if True include redundant indices
    """
    upper = itertools.count() if n_max is None else inclusive(n_max)
    for n in upper:
        indices = index_traits(n, l_max if l_max is None else min(n, l_max))
        for l in indices.iter_l(n):
            for m in inclusive(0, l):
                yield ZernikeIndex(n, l, m)

            # negative m indices go AFTER positive because we rely on the wrapping-around of negative indices
            # to make these work they are actually at the end of the sequence of m values
            if redundant:
                for m in inclusive(-l, -1, 1):
                    yield ZernikeIndex(n, l, m)


def iter_indices_extended(key: Tuple[int, Tuple], max_order: int, redundant=False) -> Iterator[ZernikeIndex]:
    # pylint: disable=too-many-branches
    if isinstance(key, int):
        # Assume the caller is treating us like a vector
        n, l, m = triple_index(key)
    else:
        # Assume tuple
        n, l, m = key

    if n is None or l is None or m is None:
        # Define the range of n values
        if n is None:
            n_iter = inclusive(0 if l is None else l, max_order)
        else:
            n_iter = inclusive(n, n)

        for n_idx in n_iter:
            # Define the range of l values
            if l is None:
                l_iter = inclusive(0, n_idx)
            else:
                l_iter = inclusive(l, l)

            for l_idx in l_iter:
                if not even(n_idx - l_idx):
                    continue

                # Define the range of m values
                if m is None:
                    m_start = -l_idx if redundant else 0
                    m_iter = inclusive(m_start, l_idx)
                else:
                    m_iter = inclusive(m, m)

                for m_idx in m_iter:
                    yield ZernikeIndex(n_idx, l_idx, m_idx)

    else:
        # All direct indices
        yield ZernikeIndex(n, l, m)


def triple_index(linear: int, redundant=True) -> ZernikeIndex:
    """Get the triple index from the given lexicographic index"""
    entry = None
    for entry, _ in zip(iter_indices(redundant=redundant), range(linear + 1)):
        pass
    return entry  # pylint: disable=inconsistent-return-statements


@functools.lru_cache(maxsize=256)
def linear_index(index: base_moments.Index, redundant=True) -> int:
    """Given a triple of zernike function indices this will return a lexicographically ordered
    integer"""
    # pylint: disable=inconsistent-return-statements
    if not redundant and index[2] < 0:
        raise ValueError('m value cannot be negative')

    for linear, triple in enumerate(iter_indices(redundant=redundant)):
        if triple == index:
            return linear

    raise RuntimeError('Should never reach here!')


def _domain_check(positions: np.array):
    """Check that a given set of positions are within the domain for which Zernike moments are
    defined i.e. that |r| <= 1."""
    if positions.dtype == object:
        return

    for idx, pos in enumerate(positions):
        if np.dot(pos, pos) > 1.:
            _LOGGER.warning(
                'Delta function %i is outside of the domain of Zernike basis functions '
                'with are defined within the unit ball', idx
            )


def index_traits(n, l) -> sph.IndexTraits:
    return sph.IndexTraits(n, l, l_le_n=True, n_minus_l_even=True)


def zernike_poly(n: int, l: int, m: int, pt, condon_shortley=True):
    """Evaluate the 3D Zernike polynomial at the given spherical coordinates"""
    rho = pt[0]
    theta = pt[1]  # need to be [0, pi]
    # I'm not sure why but there seems to be a pi/2 phase difference when using sph_harm, so just include it here
    phi = pt[2] + np.pi / 2

    theta = np.where(theta < 0, np.pi + theta, theta)
    phi = np.where(phi < 0, np.pi + phi, phi)

    radial = r_nl(n, l, rho)
    # Note that scipy uses the opposite of the physics convention, in scipy
    #   theta = azimuth (longitudinal) coord
    #   phi = polar (colatitudinal) coord
    # while we use the opposite convention
    angular = special.sph_harm(m, l, phi, theta)
    if not condon_shortley:
        # scipy sph_harm incorporates the Condon-Shortley phase so if we don't want it we have to
        # cancel it out here
        angular /= (-1)**m

    return radial * angular


def r_nl(n: int, l: int, rho) -> numbers.Number:
    """Evaluate a Zernike radial polynomial at the given value of rho"""
    if not even(n - l):
        return 0.
    #
    # D = 3.
    # k = (n - l) / 2.
    #
    # normalisation = np.sqrt(2 * n + D)
    # total = np.sum([((-1)**s) * binomial_coeff(k, s) * binomial_coeff(n - s - 1. + D / 2., k) * rho**(n - 2 * s)
    #                 for s in inclusive(int(k))],
    #                axis=0)
    #
    # # Alternative method
    # # total2 = (-1)**k * binomial_coeff((n + l + D) / 2 - 1, k) * rho**l * \
    # #          special.hyp2f1(-k, (n + l + D) / 2., l + D / 2., rho**2)
    #
    # return normalisation * total
    # #
    k = int((n - l) / 2)
    return rho**l * sum(q_kl_nu(k, l, nu) * rho**(2 * nu) for nu in inclusive(k))



# -*- coding: utf-8 -*-
import random

import numpy as np
import pytest

import milad
from milad import invariants, generate, dat

READ_MAX = 128  # The maximum number of invariants to read from the file

# pylint: disable=redefined-outer-name, invalid-name


@pytest.fixture(scope='session')
def geometric_invariants():
    """Get geometric moment invariants"""
    invs = milad.invariants.read(filename=invariants.GEOMETRIC_INVARIANTS, read_max=READ_MAX)
    return invs


@pytest.fixture(scope='session')
def complex_invariants():
    """Get complex moment invariants"""
    invs = milad.invariants.read(filename=invariants.COMPLEX_INVARIANTS)
    return invs


@pytest.fixture(autouse=True)
def set_random_seed():
    do_set_random_seeds()


@pytest.fixture()
def save_figures():
    """Return True if you want to save figures from tests"""
    return False


@pytest.fixture()
def chiral_tetrahedra():
    """Create chiral structures that cannot be distinguished by bispectrum from
    https://link.aps.org/doi/10.1103/PhysRevLett.125.166001
    """
    return generate.chiral_tetrahedra()


@pytest.fixture(scope='session')
def training_data():
    do_set_random_seeds()
    return create_fake_training_data()


@pytest.fixture(scope='session')
def descriptor(complex_invariants):
    return create_descriptor(complex_invariants)


@pytest.fixture(scope='session')
def fingerprint_set(descriptor, training_data):
    return dat.create_fingerprint_set(descriptor, training_data, get_derivatives=True)


def create_descriptor(invs, _species=('Si',)):
    cutoff = 1.

    specie_numbers = (14,)
    return milad.descriptor(
        species=dict(map=dict(numbers=specie_numbers, range=(0.5, 1.5))),
        features=dict(type=milad.functions.WeightedDelta, map_species_to='WEIGHT'),
        moments_calculator=milad.zernike.ZernikeMomentsCalculator(invs.max_order),
        invs=invs,
        cutoff=cutoff,
        apply_cutoff=True,
        smooth_cutoff=True
    )


def create_fake_training_data(num_systems=10, max_atoms=18, species=('Si',)):
    import ase
    from ase.calculators.singlepoint import SinglePointCalculator

    training_data = []
    for _ in range(num_systems):
        natoms = random.randint(1, max_atoms)

        atoms = ase.Atoms(positions=generate.random_points_in_sphere(natoms), symbols=random.choices(species, k=natoms))
        calc = SinglePointCalculator(atoms, energy=-100 * random.random(), forces=np.random.rand(natoms, 3))
        atoms.calc = calc

        training_data.append(atoms)

    return training_data


def do_set_random_seeds():
    random.seed(1234)
    np.random.seed(1234)



# -*- coding: utf-8 -*-
import collections
from typing import List, Sequence, Sized, Optional, Union, Tuple

import ase
import matplotlib.pyplot as plt
import numpy as np

__all__ = 'SystemInfo', 'FingerprintSet', 'create_fingerprint_set'

SystemInfo = collections.namedtuple('SystemInfo', 'atoms fingerprints derivatives')


class FingerprintSet:
    """Container to store and analyse a set of fingerprints"""

    def __init__(self, fingerprint_length: int):
        self._fingerprint_length = fingerprint_length
        self._system_info: List[SystemInfo] = []

    def __len__(self) -> int:
        """Return the number of systems (not, this is NOT the number of environments)"""
        return len(self._system_info)

    @property
    def fingerprint_len(self) -> int:
        """Get the length of the fingerprint vector"""
        return self._fingerprint_length

    @property
    def total_environments(self) -> int:
        """Get the total number of environments summing up over all of the systems"""
        return sum(len(info.atoms) for info in self._system_info)

    @property
    def fingerprints(self) -> Sequence:
        """Access all the fingerprints"""
        return tuple(info.fingerprints for info in self._system_info)

    def has_all_derivatives(self):
        try:
            self.fingerprint_derivatives.index(None)
            return False
        except ValueError:
            # None not found
            return True

    @property
    def fingerprint_derivatives(self) -> Sequence:
        """Access all the fingerprint derivatives"""
        return tuple(info.derivatives for info in self._system_info)

    @property
    def sizes(self):
        return tuple(len(info.atoms) for info in self._system_info)

    @property
    def systems(self) -> List[ase.Atoms]:
        return [info.atoms for info in self._system_info]

    def get_potential_energies(self, normalise=True) -> tuple:
        """Get the potential energies of all systems in the set"""
        if normalise:
            return tuple(info.atoms.get_potential_energy() / len(info.atoms) for info in self._system_info)

        return tuple(info.atoms.get_potential_energy() for info in self._system_info)

    def has_all_forces(self) -> bool:
        try:
            self.get_forces().index(None)
            return False
        except ValueError:
            # None not found
            return True

    def get_forces(self) -> Tuple[np.ndarray]:
        all_forces = []
        for info in self._system_info:
            try:
                forces = info.atoms.get_array('force')
            except KeyError:
                forces = info.atoms.get_forces()
            all_forces.append(forces)

        return tuple(all_forces)

    def add_system(self, system: ase.Atoms, fingerprints: Sized, derivatives: Optional[Sized] = None):
        """Add a system along with the fingerprints for all its environments"""
        if len(system) == 0:
            raise ValueError('Cannot add an empty atoms object')

        if len(system) != len(fingerprints):
            raise ValueError(
                'There must be as many fingerprints as there are atoms, ' \
                'got {} atoms and {} environments'.format(len(system), len(fingerprints)))
        if derivatives is not None and len(derivatives) != len(system):
            raise ValueError('Number of derivatives must match the number of atoms (environments) in the system')

        derivatives = np.array(derivatives) if derivatives is not None else None
        self._system_info.append(SystemInfo(system, np.array(fingerprints), derivatives))

    def systemwise_sum(self, values, normalise=True):
        """Given a vector of values, one per environment, this will sum up the values for each atomic system
        and return the result as a container whose length is the same as the number of systems

        The sum can optionally be normalised by the number of atoms in each system.
        """
        out = []
        idx = 0
        for info in self._system_info:
            natoms = len(info.atoms)
            summed = sum(values[idx:idx + natoms])
            if normalise:
                summed = summed / natoms
            out.append(summed)
            idx += natoms

        return out

    def plot_environments(self):
        """Create a plot of the environments.  Returns the matplotlib figure when can then be .show()n"""
        fig, axes = plt.subplots(figsize=(16, 5))

        for envs in self.fingerprints:
            for env in envs:
                axes.plot(tuple(range(len(env))), env, linewidth=1.0)

        return fig

    def plot_energies(self):
        fig, axes = plt.subplots(figsize=(16, 5))

        energies = self.get_potential_energies(normalise=True)
        axes.hist(energies, 50, log=True)
        axes.set_xlabel('Energy')
        axes.set_ylabel('Number')

        return fig

    def split(self, split_point: Union[float, int]) -> Tuple['FingerprintSet', 'FingerprintSet']:
        """Split this set into two.  This can be used for creating a training and validation set.

        The split point can be an integer, in which case it is treated as the index of the split point,
        or a float in which case is it treated as a ratio.
        """
        if isinstance(split_point, float):
            split_point = int(round(len(self) * split_point))
        elif not isinstance(split_point, int):
            raise TypeError('split_point must be integer or float, got {}'.format(split_point.__class__.__name__))

        # Split into two halves
        one = FingerprintSet(self.fingerprint_len)
        for info in self._system_info[:split_point]:
            one.add_system(*info)

        two = FingerprintSet(self.fingerprint_len)
        for info in self._system_info[split_point:]:
            two.add_system(*info)

        return one, two


def create_fingerprint_set(
    descriptor: 'milad.Descriptor', systems: Sequence[ase.Atoms], get_derivatives=False
) -> FingerprintSet:
    """Given a descriptor and a sequence of ase.Atoms objects this will create a fignerprint set"""
    # pylint: disable=too-many-locals
    from milad.play import asetools

    # WARNING: The calculation of derivatives only takes into account positional degrees of freedom (not species) and
    # makes assumptions about the shape of the derivatives tensor
    fp_length = descriptor.fingerprint_len

    fingerprints = FingerprintSet(descriptor.fingerprint_len)
    for system in systems:
        fps = []
        natoms = len(system)
        derivs = np.zeros((natoms, natoms, 3, fp_length)) if get_derivatives else None

        for my_idx, env in asetools.extract_environments(
            system, cutoff=descriptor.cutoff, yield_indices=True, include_central_atom=False
        ):
            milad_env = asetools.ase2milad(env)
            if get_derivatives:
                fingerprint, jacobian = descriptor(milad_env, jacobian=True)

                # The yielded environment has this array that allows us to map back on to the index in the original
                # structure
                orig_indices = env.get_array('orig_indices', copy=False)
                for i in range(len(env)):
                    neighbour_idx = orig_indices[i]
                    # Add the derivatives as the same neighbour may contribute more than once
                    derivs[my_idx, neighbour_idx, :, :] += jacobian[:, i * 3:(i + 1) * 3].T  # pylint: disable=unsupported-assignment-operation

                fps.append(fingerprint)
            else:
                fps.append(descriptor(milad_env))

        fingerprints.add_system(system, fps, derivatives=derivs)

    return fingerprints



# -*- coding: utf-8 -*-
import itertools
from typing import Tuple

import numpy as np
from numpy import linalg

# pylint: disable=invalid-name


def to_real(vec: np.array) -> np.array:
    """Convert a given numpy vector containing complex numbers to one twice as long containing
    only real numbers where the first half contains the real and the second half the imaginary parts
    """
    view = vec.view('(2,)float')
    real = view.reshape(view.shape[0] * view.shape[1])
    return real


def to_complex(vec: np.array) -> np.array:
    """Given a vector of real numbers convert it to one half the size containing complex numbers
    where the first half of the original vector is treated as the real parts while the second half
    is used for the imaginary"""
    half_size = int(vec.size / 2)
    reshaped = vec.reshape((half_size, 2))
    view = reshaped.view(dtype=complex).reshape(half_size)

    return view


def even(val: int) -> bool:
    """Test if an integer is event.  Returns True if so."""
    return (val % 2) == 0


def odd(val: int) -> bool:
    """Test if an integer is odd.  Returns True if so."""
    return (val % 2) != 0


def cholesky(gram: np.ndarray) -> np.array:
    """Find Cholesky decomposition of the passed Gram matrix.  If this fails the algorithm
    will attempt to force it to be positive definite

    A Python/Numpy port of John D'Errico's `nearestSPD` MATLAB code [1], which credits [2].

    [1] https://www.mathworks.com/matlabcentral/fileexchange/42885-nearestspd

    [2] N.J. Higham, "Computing a nearest symmetric positive semidefinite matrix" (1988):
    https://doi.org/10.1016/0024-3795(88)90223-6
    """

    # pylint: disable=invalid-name
    try:
        return np.linalg.cholesky(gram)
    except np.linalg.LinAlgError:
        pass

    B = (gram + gram.T) / 2
    _, s, V = np.linalg.svd(B)

    H = np.dot(V.T, np.dot(np.diag(s), V))

    A2 = (B + H) / 2
    A3 = (A2 + A2.T) / 2

    spacing = np.spacing(np.linalg.norm(gram))
    # The above is different from [1]. It appears that MATLAB's `chol` Cholesky
    # decomposition will accept matrixes with exactly 0-eigenvalue, whereas
    # Numpy's will not. So where [1] uses `eps(mineig)` (where `eps` is Matlab
    # for `np.spacing`), we use the above definition. CAVEAT: our `spacing`
    # will be much larger than [1]'s `eps(mineig)`, since `mineig` is usually on
    # the order of 1e-16, and `eps(1e-16)` is on the order of 1e-34, whereas
    # `spacing` will, for Gaussian random matrixes of small dimension, be on
    # othe order of 1e-16. In practice, both ways converge, as the unit test
    # below suggests.
    I = np.eye(gram.shape[0])
    k = 1
    while True:
        try:
            return np.linalg.cholesky(A3)
        except np.linalg.LinAlgError:
            pass

        mineig = np.min(np.real(np.linalg.eigvals(A3)))
        A3 += I * (-mineig * k**2 + spacing)
        k += 1


def pol2cart(r: float, theta: float) -> Tuple[float, float]:
    """
    Convert from polar to Cartesian
    """
    # pylint: disable=invalid-name
    z = r * np.exp(1j * theta)
    x, y = z.real, z.imag

    return x, y


def sph2cart(vec) -> np.array:
    """Convert spherical coordinates to Cartesian using 'physicists convention'"""
    # pylint: disable=invalid-name
    r = vec[0]
    theta = vec[1]
    phi = vec[2]

    x = r * np.sin(theta) * np.cos(phi)
    y = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(theta)
    return np.array([x, y, z])


def cart2sph(vec) -> np.array:
    x = vec[0]
    y = vec[1]
    z = vec[2]

    r = np.sqrt(x**2 + y**2 + z**2)
    theta = np.arccos(z / r)
    phi = np.arctan2(y, x)

    # hxy = np.hypot(x, y)
    # r = np.hypot(hxy, z)
    #
    # theta = np.arctan2(hxy, z)
    # phi = np.arctan2(y, x)

    # Wrap phi to the range [0, 2pi]
    phi2 = np.where(phi < 0, 2 * np.pi + phi, phi)
    return np.array([r, theta, phi2])


def krank(array: np.ndarray) -> int:
    """Compute the k-rank (sometimes called spark) of a matrix

    See: https://en.wikipedia.org/wiki/Spark_(mathematics)
    """
    # This is the maximum possible k-rank
    rank = linalg.matrix_rank(array)
    ncols = array.shape[1]
    cols = tuple(range(ncols))

    for k in range(2, rank):
        for comb in itertools.combinations(cols, k):
            r = linalg.matrix_rank(array[:, comb])
            if r < k:
                return k

    return rank



# -*- coding: utf-8 -*-
import math

import ase.build
import ase.io
import ase.neighborlist
import matplotlib.pyplot as plt
import numpy as np

import milad
from milad.play import asetools

# pylint: disable=invalid-name


def test_multiple_species(geometric_invariants, request, save_figures):
    molecule = ase.build.molecule('CH3CH2OH')
    species = ['C', 'H', 'O']
    symbols = molecule.symbols
    num_atoms = len(symbols)

    fp = asetools.MiladFingerprint(species, geometric_invariants, sigmas={'C': 1., 'H': 0.5, 'O': 0.7})

    fingerprints = fp.create(molecule)

    labels = []
    for i in range(num_atoms):
        labels.append('{}$_{}$'.format(i, symbols[i]))

    assert (fingerprints[3] - fp.create_single(molecule, 3)).max() == 0

    fig, axes = plt.subplots()
    axes.set_yscale('log')
    milad.plot.plot_multiple_invariants(fingerprints, axes, labels=labels)
    axes.set_title('C$_2$H$_6$O')
    fig.legend()
    if save_figures:
        fig.savefig('{}.pdf'.format(request.node.name))


def test_multiple_species_split(geometric_invariants, request, save_figures):
    molecule = ase.build.molecule('CH3CH2OH')
    species = ['C', 'H', 'O']

    fp = asetools.MiladFingerprint(
        species, geometric_invariants, sigmas={
            'C': 0.4,
            'H': 0.3,
            'O': 0.5
        }, split_specie_pairs=True, cutoffs=100.
    )

    fingerprint = fp.create_single(molecule, 2)

    fig, axes = plt.subplots()
    axes.set_yscale('log')
    axes.set_title(str(molecule))
    milad.plot.plot_multiple_invariants(fingerprint, axes, labels=species)
    fig.legend()
    if save_figures:
        fig.savefig('{}.pdf'.format(request.node.name))


def test_generate_environments_molecule():
    molecule = ase.build.molecule('CH3CH2OH')

    envs = list(asetools.extract_environments(molecule, cutoff=4.))
    assert len(envs) == len(molecule)


def test_generate_environments_solid():
    # pylint: disable=too-many-nested-blocks, too-many-locals
    lattice_param = 2.5
    cutoff = 4.

    diamond = ase.build.bulk('C', 'hcp', a=lattice_param)
    envs = list(asetools.extract_environments(diamond, cutoff=cutoff))

    lattice_max = int(math.ceil(lattice_param / cutoff) + 1)

    cutoff_sq = cutoff * cutoff
    # lattice params
    a, b, c = diamond.cell[0], diamond.cell[1], diamond.cell[2]

    found_positions = []
    for i, central in enumerate(diamond.positions):
        env_positions = []
        for _j, other in enumerate(diamond.positions):
            r_ij = other - central

            for l in range(-lattice_max, lattice_max + 1):
                for m in range(-lattice_max, lattice_max + 1):
                    for n in range(-lattice_max, lattice_max + 1):
                        # if l == m == n == 0:
                        #     continue

                        dr = r_ij + l * a + m * b + n * c
                        if np.dot(dr, dr) < cutoff_sq:
                            env_positions.append(dr)

        known_positions = envs[i].positions
        assert len(env_positions) == len(known_positions)
        positions = np.array(env_positions)
        found_positions.append(positions)

        # Check the list of distances matches at least
        assert sorted(np.sum(positions * positions, axis=1)) == \
               sorted(np.sum(known_positions * known_positions, axis=1))

    assert len(envs) == len(found_positions)


def test_extract_environments():
    num_atoms = 5
    cutoff = 0.7
    system = ase.Atoms(
        positions=[[0., 0., 0.]] * num_atoms, numbers=[6] * num_atoms, cell=[1, 1, 1, 47, 97, 121], pbc=True
    )
    system.rattle(stdev=156)
    system.wrap()

    nlist = ase.neighborlist.NeighborList([cutoff / 2.] * num_atoms, skin=0., self_interaction=False, bothways=True)
    nlist.update(system)

    for i, env in tuple(
        asetools.extract_environments(system, cutoff=cutoff, yield_indices=True, include_central_atom=False)
    ):
        indices, offsets = nlist.get_neighbors(i)
        pos = system.positions[i]

        Rs = np.array([
            system.positions[j] + np.dot(offset, system.cell) - pos for (j, offset) in zip(indices, offsets)
        ])

        indices = sorted(indices)
        env_indices = sorted(env.get_array('orig_indices'))
        assert env_indices == indices

        dots = np.array(sorted(np.sum(Rs * Rs, axis=1)))
        env_dots = np.array(sorted(np.sum(env.positions * env.positions, axis=1)))

        assert np.allclose(dots, env_dots)



# -*- coding: utf-8 -*-

try:
    import amp  # pylint: disable=unused-import
except ImportError:
    __all__ = tuple()
else:
    import argparse
    import collections
    from typing import Dict, Any

    from amp import utilities
    import ase
    from ase.calculators.calculator import Parameters

    from milad.play import asetools
    from . import interfaces

    __all__ = ('AmpDescriptor',)

    class AmpDescriptor:
        """This adapter allows MILAD descriptors to be used by the amp code:
         https://amp.readthedocs.io/en/latest/

         Functionality is not complete but the basics do work.
         """

        def __init__(self, descriptor: interfaces.Descriptor, dblabel=None):
            self._descriptor = descriptor
            self.dblabel = dblabel
            self.parent = None
            self.parameters = Parameters({'cutoff': descriptor.cutoff, 'mode': 'atom-centered'})

        @property
        def cutoff(self) -> float:
            return self._descriptor.cutoff

        def load_amp(self, filename: str, label='') -> amp.Amp:
            """Given a checkpoint filename this will load the Amp class and pass it this descriptor.

            This means that the caller needs to be sure that this descriptor is the one used for when training the
            loaded model, otherwise there will be inconsistencies.
            """
            if not label:
                label = filename.replace('.amp', '')
            return amp.Amp.load(filename, Descriptor=lambda *_args, **kwargs: self, label=label)

        def tostring(self) -> str:
            """Returns an evaluatable representation of the calculator that can
            be used to restart the calculator."""
            return self.parameters.tostring()

        def calculate_fingerprints(
            # pylint: disable=unused-argument
            self,
            images: Dict[Any, ase.Atoms],
            parallel=None,
            log=None,
            calculate_derivatives=False
        ):
            log = utilities.Logger(file=None) if log is None else log

            if self.dblabel is None:
                if hasattr(self.parent, 'dblabel'):
                    self.dblabel = self.parent.dblabel
                else:
                    self.dblabel = 'amp-data'

            log('Fingerprinting images...', tic='fp')
            if not hasattr(self, 'fingerprints'):
                # pylint: disable=attribute-defined-outside-init
                self.fingerprints = utilities.Data(
                    filename=f'{self.dblabel}-fingerprints',
                    calculator=argparse.Namespace(calculate=self._calc_fingerprint)
                )

            self.fingerprints.calculate_items(images, parallel=dict(cores=1), log=log)
            log('...fingerprints calculated.', toc='fp')

            if calculate_derivatives:
                log('Calculating fingerprint derivatives of images...', tic='derfp')
                if not hasattr(self, 'fingerprintprimes'):
                    # pylint: disable=attribute-defined-outside-init
                    self.fingerprintprimes = utilities.Data(
                        filename=f'{self.dblabel}-fingerprints-primes',
                        calculator=argparse.Namespace(calculate=self._calc_fingerprint_derivatives)
                    )  # pylint: disable=attribute-defined-outside-init

                self.fingerprintprimes.calculate_items(images, parallel=dict(cores=1), log=log)
                log('...fingerprint derivatives calculated.', toc='derfp')

        def _calc_fingerprint(self, system: ase.Atoms, _hashval):
            fingerprints = []

            for my_idx, env in asetools.extract_environments(
                system, cutoff=self._descriptor.cutoff, yield_indices=True, include_central_atom=False
            ):
                my_symbol = system.symbols[my_idx]
                milad_env = asetools.ase2milad(env)
                fingerprints.append((my_symbol, self._descriptor(milad_env)))

            return fingerprints

        def _calc_fingerprint_derivatives(self, system: ase.Atoms, _hashval):
            # pylint: disable=too-many-locals
            derivatives = collections.defaultdict(float)

            for my_idx, env in asetools.extract_environments(
                system,
                cutoff=self._descriptor.cutoff,
                yield_indices=True,
                include_central_atom=False,
            ):
                my_symbol = system.symbols[my_idx]
                milad_env = asetools.ase2milad(env)

                _, jac = self._descriptor(milad_env, jacobian=True)

                natoms = len(env)  # Number of atoms in the environment

                # The yielded environment has this array that allows us to map back on to the index in the
                # original structure
                orig_indices = env.get_array('orig_indices', copy=False)
                derivs = collections.defaultdict(float)

                for i in range(natoms):
                    neighbour_idx = orig_indices[i]
                    if neighbour_idx == my_idx:
                        continue

                    # Add to the derivatives as the same neighbour may contribute more than once
                    local_derivs = jac[:, i * 3:(i + 1) * 3].T
                    derivs[neighbour_idx] += local_derivs

                # Now copy over the derivatives to AMP format
                for neighbour_idx, derivs_ in derivs.items():
                    neighbour_symbol = system.symbols[neighbour_idx]

                    for coord in range(3):
                        # Derivative of this fingerprint wrt coordinates of central atom
                        # calculated as the negative of the derivative wrt to the positions of the neighbours i.e.
                        # -d/dr_ij.  We sum up over all neighbours.
                        deriv_idx = (my_idx, my_symbol, my_idx, my_symbol, coord)
                        derivatives[deriv_idx] += -derivs_[coord]

                        # Derivative of this fingerprint wrt coordinates of neighbouring atom
                        deriv_idx = (neighbour_idx, neighbour_symbol, my_idx, my_symbol, coord)
                        derivatives[deriv_idx] = derivs_[coord]

            return {key: val.real.tolist() for key, val in derivatives.items()}



# -*- coding: utf-8 -*-
import numpy as np

from milad import polynomials

# pylint: disable=invalid-name


# Let's create a polynomial function to be able to compare
def poly_fn(x: np.array):
    return 5 * x[0, 0, 0] * x[0, 1, 0] + 3 * x[0, 1, 0] * x[0, 1, 0] + 9


POLY = polynomials.HomogenousPolynomial(2, [5, 3], [[[0, 0, 0], [0, 1, 0]], [[0, 1, 0], [0, 1, 0]]], constant=9)


def test_polynomial():
    x = np.random.rand(2, 2, 2)
    assert poly_fn(x) == POLY(x)
    x = np.zeros((2, 2, 2))
    assert poly_fn(x) == POLY(x)

    # Check partial derivatives
    deriv = POLY.get_partial_derivative((0, 0, 0))
    assert deriv.degree == POLY.degree - 1
    assert np.all(deriv.prefactors == np.array([5.]))
    assert np.all(deriv.terms == np.array([[[0, 1, 0]]]))
    assert deriv.constant == 0.


def test_partial_derivatives():
    deriv = POLY.get_partial_derivative((0, 1, 0))
    assert deriv.degree == POLY.degree - 1
    assert np.all(deriv.prefactors == np.array([5., 6.]))
    assert np.all(deriv.terms == np.array([[[0, 0, 0]], [[0, 1, 0]]]))
    assert deriv.constant == 0.

    # Let's do the derivative again so we can check how it deals with terms whose power drops to 0
    deriv2 = deriv.get_partial_derivative((0, 1, 0))
    assert deriv2.degree == deriv.degree - 1
    assert np.all(deriv2.prefactors == np.array([]))
    assert np.all(deriv2.terms == np.array([]))
    assert deriv2.constant == 6.

    # And finally let's check what happens when we do a derivative of a constant
    deriv3 = deriv2.get_partial_derivative((0, 1, 0))
    assert deriv3.degree == deriv.degree - 1
    assert np.all(deriv3.prefactors == np.array([]))
    assert np.all(deriv3.terms == np.array([]))
    assert deriv3.constant == 0.


def test_gradient():
    grad = POLY.get_gradient()
    assert set(grad.keys()) == POLY.variables


def test_simplify():
    """Test that a polynomial will automatically simplify itself if there are commons factors in the sum"""
    # This is effectively 5 * x[1, 0, 0] + 7 * x[1, 0, 0] which should be simplified
    poly = polynomials.HomogenousPolynomial(1, [5, 7], [[[1, 0, 0]], [[1, 0, 0]]])
    assert np.all(poly.prefactors == np.array([12]))
    assert np.all(poly.terms == np.array([[[1, 0, 0]]]))



# -*- coding: utf-8 -*-
import numpy as np

try:
    import dscribe.descriptors
except ImportError:
    __all__ = tuple()
else:
    from milad.play import asetools
    from milad import atomic
    from . import interfaces

    __all__ = ('DScribeDescriptor',)

    class DScribeDescriptor(interfaces.Descriptor):

        def __init__(self, dscribe_descriptor: dscribe.descriptors.Descriptor):
            super().__init__()
            self._dscribe = dscribe_descriptor

        @property
        def fingerprint_len(self) -> int:
            return self._dscribe.get_number_of_features()

        @property
        def cutoff(self) -> float:
            try:
                return getattr(self._dscribe, '_rcut')
            except AttributeError:
                return float('inf')

        def evaluate(self, atoms: atomic.AtomsCollection, *, get_jacobian=False):
            ase_atoms = asetools.milad2ase(atoms)

            if get_jacobian:
                natoms = atoms.num_atoms
                try:
                    res = self._dscribe.derivatives(ase_atoms, positions=[0])
                except AttributeError:
                    raise RuntimeError('Calculation of derivatives not supported')
                else:
                    jac = np.zeros((self.fingerprint_len, natoms * 4))
                    jac[:, :3 * natoms] = res[0].reshape(natoms * 3, self.fingerprint_len).T
                    return res[1][0], jac

            return self._dscribe.create(ase_atoms, positions=[0])[0]



# -*- coding: utf-8 -*-
import random

import matplotlib.pyplot as plt
import numpy as np
import sympy

import milad
from milad import zernike
from milad import functions
from milad import generate
from milad import geometric


def test_invariant_single_mass(geometric_invariants, request, save_figures):
    num_invariants = 20
    num_masses = 20

    origin = np.zeros((1, 3))

    fig, axes = plt.subplots()

    for i in range(num_masses):
        mass = 0. + (0.1 * i)

        moments = geometric.from_gaussians(geometric_invariants.max_order, origin, sigmas=0.4, weights=mass)
        invariants = geometric_invariants[:num_invariants](moments)

        milad.plot.plot_invariants(invariants, axes, label='mass={}'.format(mass))

    fig.legend()
    if save_figures:
        fig.savefig('{}.pdf'.format(request.node.name))


def test_invariant_two_weights(geometric_invariants, request, save_figures):
    num_invariants = 64
    num_weights = 11

    positions = np.array(((-2., 0., 0.), (2., 0., 0.)))

    fig, axes = plt.subplots()

    for i in range(num_weights):
        mass = 0. + (0.1 * i)

        moments = geometric.from_gaussians(geometric_invariants.max_order, positions, sigmas=0.4, weights=(1., mass))
        invariants = geometric_invariants[:num_invariants](moments)

        milad.plot.plot_invariants(invariants, axes, label=f'$w={mass:.1f}$')

    axes.set_yscale('log')
    axes.set_title('Varying mass')
    fig.legend()
    if save_figures:
        fig.savefig('{}.pdf'.format(request.node.name))


def test_invariant_derivative(geometric_invariants):
    # Take a random invariant and make sure that derivatives are getting calculated accurately
    invariant = random.choice(geometric_invariants)

    # Symbols for moments
    m = sympy.IndexedBase('m')  # pylint: disable=invalid-name
    phi = invariant.apply(m)  # Analytic expression for moments
    derivatives = invariant.get_gradient()

    for indices, entry in derivatives.items():
        dm = m[indices]  # pylint: disable=invalid-name
        dphi_dm_analytic = phi.diff(dm)
        dphi_dm_calculated = entry(m)

        assert dphi_dm_calculated == dphi_dm_analytic


def test_invariants_function(geometric_invariants):
    num_points = 10
    max_order = 10

    moments_fn = geometric.GeometricMomentsCalculator(max_order)

    pts = generate.random_points_in_sphere(num_points)
    env = functions.Features(*map(functions.WeightedDelta, pts))
    moments = moments_fn(env)
    phi = geometric_invariants(moments)

    # The 1st moment is always the total mass
    assert phi[0] == num_points

    # Now try the same thing using chain
    combined_fn = functions.Chain(moments_fn, geometric_invariants)
    phi2, _jacobian = combined_fn(env, jacobian=True)

    assert np.all(phi == phi2)


def test_invariants_derivatives_correctness(complex_invariants):
    complex_invariants = complex_invariants[:10]
    # The moments
    m = sympy.IndexedBase('m', complex=True)  # pylint: disable=invalid-name
    # Fill with symbols
    zernike_moms = milad.ZernikeMoments.from_indexed(m, complex_invariants.max_order, dtype=object)

    phi, jac = complex_invariants(zernike_moms, jacobian=True)

    for i, inv in enumerate(phi):
        for j, (_, mom) in enumerate(zernike_moms.iter(redundant=True)):
            # Perform symbolic derivative
            diff = sympy.diff(inv, mom).expand()

            # Get derivative from the Jacobian
            from_jac = jac[i, j]
            if not isinstance(from_jac, (int, np.integer)):
                from_jac = from_jac.expand()

            # Compare
            difference = diff - from_jac
            if not diff == from_jac:
                # If they differ, check that it's by a meaningful amount
                coeffs = np.array(tuple(difference.as_coefficients_dict().values()))
                np.testing.assert_array_almost_equal(coeffs, 0., decimal=10)


def test_against_chiral_tetrahedra(complex_invariants, chiral_tetrahedra):
    minus, plus = chiral_tetrahedra

    minus_moms = zernike.from_deltas(complex_invariants.max_order, minus)
    plus_moms = zernike.from_deltas(complex_invariants.max_order, plus)

    minus_phi = complex_invariants(minus_moms)
    plus_phi = complex_invariants(plus_moms)

    assert not np.allclose(minus_phi, plus_phi)
